from django.middleware.csrf import CsrfViewMiddleware
from django.utils.deprecation import MiddlewareMixin

from . import mixins


class CSRFCheck(CsrfViewMiddleware):
    def _reject(self, request, reason):
        # Return the failure reason instead of an HttpResponse
        return reason


class TSSOMiddleware(mixins.TSSOAuthenticationMixin, MiddlewareMixin):
    """Django authentication middleware implementing TSSO"""

    def process_request(self, request):
        user = getattr(request, 'user', None)
        if user and not user.is_anonymous:
            return
        pair = self._authenticate(request)
        if not pair:
            return
        user, auth = pair
        if user:
            request.user = user
            request.auth = auth
            # Ignore CSRF for TSSO authorized requests
            request.csrf_processing_done = True



# Generated by Django 5.0.3 on 2024-03-27 14:28

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="SSOToken",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "backend",
                    models.CharField(
                        help_text="The backend id for the token",
                        max_length=80,
                        verbose_name="Backend",
                    ),
                ),
                (
                    "token",
                    models.CharField(
                        help_text="The backend-specific token",
                        max_length=255,
                        verbose_name="Token",
                    ),
                ),
                (
                    "ctime",
                    models.DateTimeField(
                        auto_now_add=True,
                        db_index=True,
                        help_text="When the instance has been created",
                        verbose_name="Created",
                    ),
                ),
                (
                    "mtime",
                    models.DateTimeField(
                        auto_now=True,
                        db_index=True,
                        help_text="When the instance has been modified",
                        verbose_name="Modified",
                    ),
                ),
                (
                    "etime",
                    models.DateTimeField(
                        db_index=True,
                        editable=False,
                        help_text="When the instance should be expired, or has been expired",
                        null=True,
                        verbose_name="Expired",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        help_text="User to whom the token belongs",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="sso_tokens",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="User",
                    ),
                ),
            ],
            options={
                "verbose_name": "SSO Token",
                "verbose_name_plural": "SSO Tokens",
                "unique_together": {("backend", "token")},
            },
        ),
    ]



"""
Mixins providing TSSO Authorization
"""

import logging

from django.conf import settings
from django.utils.translation import gettext_lazy as _
from django.utils import timezone
from django.core import exceptions
from social_django.utils import load_strategy

from .models import SSOToken


logger = logging.getLogger(__name__)

# Header encoding (see RFC5987), C&P from the REST framework
HTTP_HEADER_ENCODING = 'iso-8859-1'


def _get_authorization_header(request):
    """
    Return request's 'Authorization:' header, as a bytestring.

    Hide some test client ickyness where the header can be unicode.
    """
    # C&P from the REST framework
    auth = request.META.get('HTTP_AUTHORIZATION', b'')
    if isinstance(auth, str):
        # Work around django test client oddness
        auth = auth.encode(HTTP_HEADER_ENCODING)
    return auth


class TSSOAuthenticationMixin:
    """
    Base Transparent SSO authentication mixin

    The client can authenticate itself by the token got from
    the known OAuth2 backend. It sends a token gor from the
    foreign system and identifies the system itself using
    request headers.

    The request headers should allow to identify the backend, and
    the token specifially for this backend.

    The Authentication header should look like:

    Authentication: SSO <backend-name>:Bearer:<backend-specific-token>
    """

    keyword = getattr(settings, 'SSSO_KEYWORD', 'SSO')

    def _authenticate(self, request):
        """Authenticates a request"""
        auth = _get_authorization_header(request).decode(errors='replace').split(' ', 1)

        if not auth:
            return None

        if len(auth) < 2:
            return None

        scheme, token = auth
        if scheme.lower() != self._authenticate_header(request).lower():
            return None

        auth = token.split(getattr(settings, 'SSO_TOKEN_SEPARATOR', ':'), 2)
        if len(auth) < 3:
            return None

        backend, token_type, token = auth
        return self._authenticate_credentials(backend, token_type, token, request=request)

    def _authenticate_credentials(self, backend_name, token_type, token, request=None):
        """Authenticates the request using extracted credentials"""

        model = self._get_model(request)
        sso = model.objects.select_related('user').filter(backend=backend_name, token=token).first()

        if sso and not sso.is_expired:
            if not sso.user.is_active:
                raise exceptions.PermissionDenied(_('User inactive or deleted.'))
            return (sso.user, sso)

        strategy = load_strategy()
        try:
            backend = strategy.get_backend(backend_name)
        except Exception as ex:
            raise exceptions.PermissionDenied(_('Wrong backend key %s: %s') % (backend_name, ex))
        try:
            user = backend.do_auth(token, token_type=token_type)
        except Exception as ex:
            raise exceptions.PermissionDenied(
                _('Token is not authorized: %s:%s:%s: %s') % (backend_name, token_type, token, ex)
            )

        if not user:
            raise exceptions.PermissionDenied(
                _('Token does not identify user: %s:%s:%s') % (backend_name, token_type, token)
            )

        if not user.is_active:
            raise exceptions.PermissionDenied(_('User inactive or deleted.'))

        eperiod = getattr(settings, 'TSSO_EXPIRATION_PERIOD', 600)
        etime = timezone.now() + timezone.timedelta(seconds=eperiod)

        if sso:
            if sso.user != user:
                logger.warning('The user has been changed for the token since last check')
                sso.user = user
            sso.etime = etime
            sso.save(update_fields=['user', 'etime'])
        else:
            sso = model.objects.create(backend=backend_name, token=token, user=user, etime=etime)
        return (sso.user, sso)

    def _authenticate_header(self, request):
        """Returns the authentication-specific header"""
        return self.keyword

    def _get_model(self, request):
        """Returns the model to store SSO tokens locally"""
        return SSOToken



"""
Provides various authentication policies.
"""

from django.core import exceptions as dex
from rest_framework import exceptions as rex
from rest_framework.authentication import BaseAuthentication

from tsso.mixins import TSSOAuthenticationMixin


class TSSOAuthentication(TSSOAuthenticationMixin, BaseAuthentication):
    """Implementing REST-specific"""

    def authenticate(self, request):
        """Override to make job"""
        try:
            return self._authenticate(request)
        except dex.PermissionDenied as ex:
            raise rex.AuthenticationFailed(*ex.args)

    def authenticate_header(self, request):
        """Override to make job"""
        return self._authenticate_header(request)



"""
Models
"""

from django.conf import settings
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _


class SSOToken(models.Model):
    """
    SSOToken is a token provided by the client to log in using foreign OAuth2 server.

    When the client provides a token to log in, the application checks it's presence
    in the database and its TTL and other parameters.
    """
    backend = models.CharField(
        max_length=80,
        verbose_name=_('Backend'), help_text=_('The backend id for the token')
    )
    token = models.CharField(
        max_length=255,
        verbose_name=_('Token'), help_text=_('The backend-specific token')
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE, related_name='sso_tokens',
        verbose_name=_('User'), help_text=_('User to whom the token belongs')
    )
    ctime = models.DateTimeField(
        auto_now_add=True, db_index=True,
        editable=False,
        verbose_name=_('Created'), help_text=_('When the instance has been created')
    )
    mtime = models.DateTimeField(
        auto_now=True, db_index=True,
        editable=False,
        verbose_name=_('Modified'), help_text=_('When the instance has been modified')
    )
    etime = models.DateTimeField(
        null=True,
        db_index=True,
        editable=False,
        verbose_name=_('Expired'), help_text=_('When the instance should be expired, or has been expired')
    )

    def is_expired(self):
        return self.etime and self.etime < timezone.now()
    is_expired.short_description = _('Is Expired')

    is_expired = property(is_expired)

    def __str__(self):
        """String representation"""
        return '%s @ %s [%s]' % (self.user, self.backend, '%s...%s' % (self.token[:3], self.token[-3:]))

    class Meta:
        verbose_name = _('SSO Token')
        verbose_name_plural = _('SSO Tokens')
        unique_together = (
            ('backend', 'token'),
        )



"""
Admin interface
"""
from admin_auto_filters.filters import AutocompleteFilterFactory
from django.contrib import admin
from django.utils.translation import gettext_lazy as _

from .models import SSOToken


@admin.register(SSOToken)
class SSOTokenAdmin(admin.ModelAdmin):
    """
    Admin interface
    """
    list_display = ('id', 'backend', 'user', 'ctime',  'mtime', 'etime')
    search_fields = ('backend', 'token',)
    list_filter = (
        AutocompleteFilterFactory(_('User'), 'user'),
    )
    autocomplete_fields = ('user',)

    def has_add_permission(self, request):
        """Restrict add function"""
        return False

    def has_change_permission(self, request, obj=None):
        """Restrict change function"""
        return False

    def has_delete_permission(self, request, obj=None):
        """Restrict delete function only for the superuser"""
        return request.user.is_superuser
